import gobject
import pygtk
pygtk.require('2.0')
import gtk
import pango
import sys
import os
import platform
from gtkeasyapp.preffile import *

##Constants
PLATFORM_UNKNOWN = 0
PLATFORM_MAEMO = 1
PLATFORM_LINUX = 2
PLATFORM_WINDOWS = 3

##Easy window Exception class
class EasyWindowException(Exception):
	def __init__(self, message):
		Exception.__init__(self, message)
		
		

##Hildon window class that implements some useful features like:
## - preferences file (using PrefDialog classes)
## - application sound (using gstreamer)
## - Hildon infos messages
## - Needed UI application method (menu...)
## - Platform detection (windows, linux, maemo)
class EasyWindow():
	##Constructor
	def __init__(self, appTitle, prefFilePath='', canHibernate=True):
		#set members
		self._platform = self.__detectPlatform()
		self._prefFile = None
		if len(prefFilePath)>0:
			self._prefFile = self.__loadPreferencesFile(prefFilePath)
		self._player = None
		self._mainContainer = None
		self.__vbox = None
		self._program = None
		if self._platform==PLATFORM_MAEMO:
			import hildon
			self._program = hildon.Program()
			self._program.__init__()
	
		#init window
		if self._platform==PLATFORM_MAEMO:
			self._window = hildon.Window()
			self._program.add_window(self._window)
			self._program.set_can_hibernate(canHibernate)
		else:
			self._window = gtk.Window()
		self._window.set_title(appTitle)
		self._window.connect("destroy", self._quit)
		self._window.set_default_size(800, 500)
		self.__addPlaftormContainer()
	
		#create menu
		menu = self._createMenu(self._window, self._prefFile)
		if self._platform==PLATFORM_WINDOWS:
			menus_bar = gtk.MenuBar()
			self.__vbox.pack_end(menus_bar, False, False)
			root_item = gtk.MenuItem(appTitle)
			root_item.set_submenu(menu)
			menus_bar.append(root_item)
		else:
			self._window.set_menu(menu)
	
		#create widgets
		self._createWidgets(self._window, self._prefFile)
	
		#show ui
		self._window.show_all()
		
		#check version
		self.__checkVersion()
		
		#startup treatment (like open splash screen...)
		gtk.gdk.threads_enter()
		self._startupTreatment(self._window, self._prefFile)
		gtk.gdk.threads_leave()
		
		#launch main gtk loop
		gtk.gdk.threads_enter()
		gtk.main()
		gtk.gdk.threads_leave()
		
	
	##Detect current platform
	def __detectPlatform(self):
		#detect plaftform
		if platform.platform().lower().find('arm')!=-1:
			return PLATFORM_MAEMO
		elif platform.platform().lower().find('linux')!=-1:
			return PLATFORM_LINUX
		elif platform.platform().lower().find('windows')!=-1:
			return PLATFORM_WINDOWS
		else:
			return PLATFORM_UNKNOWN
		
	
	##Add platform  container
	def __addPlaftormContainer(self):
		if self._platform==PLATFORM_WINDOWS:
			#add vbox container before main container in WIN case cause Hildon add default one in Maemo platform
			self.__vbox = gtk.VBox()
			self._window.add(self.__vbox)
			
		#create and add main container
		self._mainContainer = self._createMainContainer(self._window, self._prefFile)
		if self._platform==PLATFORM_MAEMO:
			self._window.add(self._mainContainer)
		else:
			self.__vbox.pack_start(self._mainContainer, True, True)
	
	##Create main container
	##@param window : main gtk window
	##@param prefFile : PrefFile object
	##@return main GTK container
	def _createMainContainer(self, window, prefFile):
		raise EasyWindowException('_createMainContainer() must be implemented')

		
	##Window object accessor
	def _getWindow(self):
		return self._window

	
	##PrefFile object accessor
	def _getPrefFile(self):
		return self._prefFile
		
		
	##Main container accessor
	def _getMainContainer(self):
		return self._mainContainer
		
		
	##Get old version (version during last launch, usualy read from preferences file)
	##@return None if no version or error occured
	def _getOldVersion(self, prefFile=None):
		raise EasyWindowException('_getOldVersion() must be implemented')

		
	##Get new version (usualy read from application global variable)
	##@return None if no version or error occured
	def _getNewVersion(self, prefFile=None):
		raise EasyWindowException('_getNewVersion() must be implemented')
		
	
	##Load preferences file
	def __loadPreferencesFile(self, filePath):
		#init
		continu = True
		test = 0
		prefFile = None
	
		while continu:
			#check if file exists
			if os.path.exists(filePath):
				#file exists, try to load it
				try:
					#increase test counter
					test += 1
					
					#create PrefFile object
					prefFile = PrefFile(filePath)
					
					#completed successfully
					continu = False
					
				except Exception, e:
					#unable to load preffile, create default one
					self._createDefaultPreferencesFile(filePath)
					
					#if after 3 tests we are unable to create prefile, exit with error
					if test>3:
						sys.exit('Unable to read or create preferences file "%s"' % (filePath))
			else:
				#file doesn't exists, create default one
				self._createDefaultPreferencesFile(filePath)
		
		return prefFile
		
	
	##Create new preferences file
	def _createDefaultPreferencesFile(self, filePath):
		raise EasyWindowException('_createDefaultPreferencesFile() must be implemented')
		
	
	##Check preferences file after new version detected (useful to add or remove item)
	def _checkPreferencesFile(self, filePath):
		raise EasyWindowException('_checkPreferencesFile() must be implemented')
		
		
	##Check version
	##If version is different, execute _newVersionDetected() method
	def __checkVersion(self):
		newVersion = self._getNewVersion(self._prefFile)
		if self._getOldVersion(self._prefFile)!=newVersion:
			#new version detected, check preferences file
			self._checkPreferencesFile(self._prefFile.getPrefFilePath())

	##Create application menu
	##@param window : main gtk window
	##@param prefFile : PrefFile object
	def _createMenu(self, window, prefFile):
		raise EasyWindowException('_createMenu() must be implemented')
		
		
	##Create widgets
	##@param window : main gtk window
	##@param prefFile : PrefFile object
	def _createWidgets(self, window, prefFile):
		raise EasyWindowException('_createWidgets() must be implemented')
		
		
	##Startup treatment (like open splash screen...)
	##@param window : main gtk window
	##@param prefFile : PrefFile object
	def _startupTreatment(self, window, prefFile):
		raise EasyWindowException('_startupTreatment() must be implemented')
		
	
	##Play specified sound
	def _playSound(self, filePath):
		#create sound player if necessary
		if self._player==None:
			#load libs
			import pygst
			pygst.require("0.10")
			import gst

			#init player
			self._player = Nonegst.element_factory_make("playbin", "player")
			fakesink = gst.element_factory_make("fakesink", "fakesink")
			self._player.set_property("video-sink", fakesink)
			bus = self.player.get_bus()
			bus.add_signal_watch()
			bus.connect("message", self.__onPlayerMessage)
	
		#Check if file exists
		if os.path.exists(soundFile):
			self.player.set_property("uri", "file://" + filepath)
			self.player.set_state(gst.STATE_PLAYING)
			
		
	##Player message handle
	##@info http://pygstdocs.berlios.de/pygst-tutorial/playbin.html
	def __onPlayerMessage(self, bus, message):
		t = message.type
		if t == gst.MESSAGE_EOS:
			#stop player at end of stream
			self._player.set_state(gst.STATE_NULL)
		elif t == gst.MESSAGE_ERROR:
			#error stdout it
			self._player.set_state(gst.STATE_NULL)
			err, debug = message.parse_error()
			print "Error: %s" % err, debug
		
		
	##Quit callback
	def _quit(self, widget, data=None):
		#before quit callback
		self._beforeQuit()
		
		#stop gtk main loop
		gtk.main_quit()
		
	
	##Before quit callback
	def _beforeQuit(self):
		raise EasyWindowException('_beforeQuit() must be implemented')
		
		
	##Bring window to front
	def bringToFront(self):
		self._window.present()
		
		
	##Return True if window is currently activated by the window manager
	##Works only in MAEMO platform
	def isActivated():
		if self._program!=None:
			return self._program.get_is_topmost()
		else:
			return True
			
			