import cairo
import gobject
import pygtk
pygtk.require('2.0')
import gtk
import time
import geometry
import math
import pango

##Kinetic exception
class VirtualWidgetException(Exception):
	def __init__(self, message):
		Exception.__init__(self, message)

##Base class for an item in Virtual Widget
##Inherited from gobject.GObject to allow returning this object in emit method
class KineticWidgetItem(gobject.GObject):
	##Constructor
	def __init__(self):
		#super constructor
		gobject.GObject.__init__(self)
	
		#members
		self.rect = geometry.Rectangle()
		self.pixmap = None
		self.pixmap_selected = None
		self.position = -1
	
	##Return item position in list
	##@return item position in list or -1 if not in list
	def getPosition():
		return self.position
	
	##Create cache pixmaps
	##@param drawable : drawable to draw in
	def draw(self, drawable):
		raise VirtualWidgetException('draw() method must be implemented')
		
	def drawSelected(self, drawable):
		raise VirtualWidgetException('drawSelected() method must be implemented')

##Image + text kinetic item
class KineticWidgetImageTextItem(KineticWidgetItem):
	def __init__(self, image, text):
		#super constructor
		KineticWidgetItem.__init__(self)
	
		#set members
		self.__pixbuf = gtk.gdk.pixbuf_new_from_file(image)
		self.__text = text
		self.__cache = None #cache text pixbuf when item is vertical
		self.__cacheSelected = None
		
		#set widget size (container will automatically adjust size)
		self.rect.width = self.__pixbuf.get_width() + 3
		self.rect.height = self.__pixbuf.get_height() + 3

	##Draw item in specified pixmap
	##@param pixmap : pixmap to draw in
	##@param drawingarea : widget where item are drawn in. Permits to have access graphic context, styles...
	##@param orientation : orientation of the item
	def draw(self, kineticPixmap):
		#init
		kineticPixmap.pangoLayout.set_text(self.__text)
		(textW, textH) = kineticPixmap.pangoLayout.get_pixel_size()
		itemSize = (self.__pixbuf.get_width() + textW , self.rect.height )
	
		#draw image
		kineticPixmap.pixmap.draw_pixbuf(None, self.__pixbuf, 0, 0, 1, 1, self.__pixbuf.get_width(), self.__pixbuf.get_height())
	
		#draw text
		if kineticPixmap.orientation==gtk.ORIENTATION_VERTICAL:
			kineticPixmap.pixmap.draw_layout(kineticPixmap.textGc, self.__pixbuf.get_width() + 2, (self.rect.width/2)-(textH/2), kineticPixmap.pangoLayout)
		elif kineticPixmap.orientation==gtk.ORIENTATION_HORIZONTAL:
			if self.__cache==None:
				#cache text image
				pixbuf = self.getTextPixbuf(kineticPixmap)
				self.__cache = self.rotatePixbuf(pixbuf, 270)
					
			#draw rotated and flipped image
			kineticPixmap.pixmap.draw_pixbuf(kineticPixmap.gc, self.__cache, 0, 0, (self.rect.height/2)-(textH/2), self.__pixbuf.get_height() + 2)
	
	##Draw selected item in specified pixmap
	##@param pixmap : pixmap to draw in
	##@param drawingarea : widget where item are drawn in. Permits to have access graphic context, styles...
	##@param orientation : orientation of the item
	def drawSelected(self, kineticPixmap):
		#init
		kineticPixmap.pangoLayout.set_text(self.__text)
		(textW, textH) = kineticPixmap.pangoLayout.get_pixel_size()
		itemSize = (self.__pixbuf.get_width() + textW , self.rect.height )
	
		#draw image
		kineticPixmap.pixmap.draw_pixbuf(None, self.__pixbuf, 0, 0, 1, 1, self.__pixbuf.get_width(), self.__pixbuf.get_height())
	
		#draw text
		if kineticPixmap.orientation==gtk.ORIENTATION_VERTICAL:
			kineticPixmap.pixmap.draw_layout(kineticPixmap.textGc, self.__pixbuf.get_width() + 2, (self.rect.width/2)-(textH/2), kineticPixmap.pangoLayout)
		elif kineticPixmap.orientation==gtk.ORIENTATION_HORIZONTAL:
			if self.__cacheSelected==None:
				#cache text image
				pixbuf = self.getTextPixbuf(kineticPixmap)
				self.__cacheSelected = self.rotatePixbuf(pixbuf, 270)
					
			#draw rotated and flipped image
			kineticPixmap.pixmap.draw_pixbuf(kineticPixmap.gc, self.__cacheSelected, 0, 0, (self.rect.height/2)-(textH/2), self.__pixbuf.get_height() + 2)
	
	'''def getLeftRotatedTextImage(self, kineticPixmap):
		#init
		(textW, textH) = kineticPixmap.pangoLayout.get_pixel_size()
	
		#create horizontal pixmap with text drawn inside
		pixmap = gtk.gdk.Pixmap(kineticPixmap.widget.window, textW, textH, -1)
		pixmap.draw_rectangle(kineticPixmap.gc, True, 0, 0, textW, textH)
		pixmap.draw_layout(kineticPixmap.textGc, 0, 0, kineticPixmap.pangoLayout)
		imageText = pixmap.get_image(0, 0, textW, textH)
		#pixbuf = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, True, 8, textW, textH)
		#pixbuf = pixbuf.get_from_image(imageText, kineticPixmap.widget.window.get_colormap(), 0, 0, 0, 0, textW, textH)
		#pixbuf.save('normal.png', 'png')
		
		#rotate horizontal pixmap
		pixmap = gtk.gdk.Pixmap(kineticPixmap.widget.window, textH, textW, -1)
		pixmap.draw_rectangle(kineticPixmap.gc, True, 0, 0, textH, textW)
		imageRotated = pixmap.get_image(0, 0, textH, textW)
		for x in range(textW):
			for y in range(textH):
				imageRotated.put_pixel(y, x, imageText.get_pixel(x, y))
		
		#flip rotated pixmap
		pixmap = gtk.gdk.Pixmap(kineticPixmap.widget.window, textH, textW, -1)
		pixmap.draw_rectangle(kineticPixmap.gc, True, 0, 0, textH, textW)
		imageFlipped = pixmap.get_image(0, 0, textH, textW)
		for x in range(textH):
			for y in range(textW):
				imageFlipped.put_pixel((textH-1)-x, y, imageRotated.get_pixel(x, y))
		
		return imageFlipped'''
	
	##Create a pixbuf that contains text un pango layout
	def getTextPixbuf(self, kineticPixmap):
		#init
		(textW, textH) = kineticPixmap.pangoLayout.get_pixel_size()
		
		#create horizontal pixmap with text drawn inside
		pixmap = gtk.gdk.Pixmap(kineticPixmap.drawable, textW, textH, -1)
		pixmap.draw_rectangle(kineticPixmap.gc, True, 0, 0, textW, textH)
		pixmap.draw_layout(kineticPixmap.textGc, 0, 0, kineticPixmap.pangoLayout)
		imageText = pixmap.get_image(0, 0, textW, textH)
		pixbuf = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, True, 8, textW, textH)
		pixbuf = pixbuf.get_from_image(imageText, kineticPixmap.drawable.get_colormap(), 0, 0, 0, 0, textW, textH)
		
		return pixbuf
	
	##Rotate specified pixbuf of specified angle
	##@param pixbuf : pixbuf to rotate
	##@param angle : angle to apply for rotation
	def rotatePixbuf(self, pixbuf, angle=90):
		#init
		rotatedPixbuf = pixbuf.rotate_simple(angle)

		return rotatedPixbuf
			
##Kinetic virtual list contains all items
class KineticVirtualList():
	##Constructor
	def __init__(self, orientation):
		self.width = 0
		self.height = 0
		self.__orientation = orientation
		self.__items = []
		self.__selectedItemIndex = -1
		self.itemSize = (0,0)
	
	##Add item
	def addItem(self, item):
		if item!=None:
			#check item size
			if self.itemSize==(0,0):
				#first inserted item, save items size
				self.itemSize = (item.rect.width, item.rect.height)
			elif self.itemSize!=(item.rect.width, item.rect.height):
				#item size is not correct
				raise VirtualWidgetException('All items must have the same size')
		
			#update item position
			item.position = len(self.__items)
		
			#store item
			self.__items.append(item)
		
			#update virtual list size
			self.__updateSize(item)
			
			#compute item position in virtual list
			self.__computeItemVirtualPosition(item)
	
	##Remove specified item from list
	##TODO
	def removeItem(self, item):
		#remove item
		#update virtual list size
		#compute items position after removing specified one
		pass
	
	##Compute item virtual position
	def __computeItemVirtualPosition(self, item):
		#get last item position
		if len(self.__items)>1:
			previousItem = self.__items[len(self.__items)-2]
		else:
			previousItem = None
		
		#compute item position
		if previousItem!=None:
			if self.__orientation==gtk.ORIENTATION_VERTICAL:
				item.rect.x = 0
				item.rect.y = previousItem.rect.y + previousItem.rect.height
			else:
				item.rect.x = previousItem.rect.x + previousItem.rect.width
				item.rect.y = 0
		else:
			item.rect.x = 0
			item.rect.y = 0
	
	##Update virtual widget size
	def __updateSize(self, item):
		if self.__orientation==gtk.ORIENTATION_VERTICAL:
			self.height += item.rect.height
		else:
			self.width += item.rect.width
		
	##Select item specified by position
	##TODO optimize search according to pixel position in list
	def selectItem(self, position):
		if self.__orientation==gtk.ORIENTATION_VERTICAL:
			self.__selectedItemIndex = -1
			for i in range(len(self.__items)):
				if self.__items[i].rect.y<=position.y and position.y<=(self.__items[i].rect.y + self.__items[i].rect.height):
					self.__selectedItemIndex = i
					break
		else:
			self.__selectedItemIndex = -1
			for i in range(len(self.__items)):
				if self.__items[i].rect.x<=position.x and position.x<=(self.__items[i].rect.x + self.__items[i].rect.width):
					self.__selectedItemIndex = i
					break
					
		#return selected item
		return self.getSelectedItem()
	
	##Get selected item
	def getSelectedItem(self):
		if self.__selectedItemIndex>=0:
			return self.__items[self.__selectedItemIndex]
		else:
			return None
	
	##Return all list items
	def getItems(self):
		return self.__items
		
##Click class contains infos about click on KineticList
class KineticClick():
	##Constructor
	def __init__(self, event):
		self.time = time.time()
		self.position = geometry.Point(event.get_coords()[0], event.get_coords()[1])
	
	##copy members
	def copy(self, click):
		self.time = click.time
		self.position = click.position
	
	##To string
	def __str__(self):
		return 'Click(time=%f pos=%s)' % (self.time, self.position)



##Base class for item pixmap
class KineticWidgetItemPixmap():
	def __init__(self, drawable, width, height, gc, textGc, pangoLayout, pangoContext, orientation):
		#set members
		self.pixmap = gtk.gdk.Pixmap(drawable, width, height, -1)
		self.gc = gc
		self.textGc = textGc
		self.pangoLayout = pangoLayout
		self.pangoContext = pangoContext
		self.drawable = drawable
		self.orientation = orientation
		
		#init pango layout
		#print pangoContext.get_font_description().get_family()
		#print pangoContext.get_font_description().get_size()
		self.pangoLayout.set_font_description(pangoContext.get_font_description())
		#font = pango.FontDescription('sans bold 12')
		#self.pangoLayout.set_font_description(font)
		#print self.pangoLayout.get_font_description()
		#print pangoContext.get_font_description().get_family()



##Default item pixmap class
class DefaultItemPixmap(KineticWidgetItemPixmap):
	def __init__(self, kineticWidget, width, height, orientation):
		#super constructor
		KineticWidgetItemPixmap.__init__(	self, 
											kineticWidget.window, 
											width, height, 
											kineticWidget.get_style().base_gc[gtk.STATE_NORMAL], 
											kineticWidget.get_style().text_gc[gtk.STATE_NORMAL], 
											kineticWidget.create_pango_layout(''),
											kineticWidget.get_pango_context(),
											orientation)

		#draw background
		self.pixmap.draw_rectangle(self.gc, True, 0, 0, width, height)
	
		#draw bottom separator
		if self.orientation==gtk.ORIENTATION_VERTICAL:
			self.pixmap.draw_line(kineticWidget.get_style().base_gc[gtk.STATE_INSENSITIVE], 0, height-1, width, height-1)
		elif self.orientation==gtk.ORIENTATION_HORIZONTAL:
			self.pixmap.draw_line(kineticWidget.get_style().base_gc[gtk.STATE_INSENSITIVE], width-1, 0, width-1, height)



##Default selected item pixmap class
class DefaultSelectedItemPixmap(KineticWidgetItemPixmap):
	def __init__(self, kineticWidget, width, height, orientation):
		#super constructor
		KineticWidgetItemPixmap.__init__(	self, 
											kineticWidget.window, 
											width, 
											height, 
											kineticWidget.get_style().base_gc[gtk.STATE_SELECTED], 
											kineticWidget.get_style().text_gc[gtk.STATE_SELECTED], 
											kineticWidget.create_pango_layout(''),
											kineticWidget.get_pango_context(),
											orientation)
		
		#draw background
		self.pixmap.draw_rectangle(self.gc, True, 0, 0, width, height)
		
		#draw bottom separator
		if self.orientation==gtk.ORIENTATION_VERTICAL:
			self.pixmap.draw_line(kineticWidget.get_style().base_gc[gtk.STATE_INSENSITIVE], 0, height-1, width, height-1)
		elif self.orientation==gtk.ORIENTATION_HORIZONTAL:
			self.pixmap.draw_line(kineticWidget.get_style().base_gc[gtk.STATE_INSENSITIVE], width-1, 0, width-1, height)


class NewKineticEffect():
	##Constructor
	##@param duration : duration of the effect
	def __init__(self, duration):
		self._startTime = 0 #t : time when effect starts
		self._duration = duration #d
		self._startValue = 0
		self._endValue = 0
		self._valueToReach = 0
		
	def _getTimeDiff(self):
		timeDiff = 0.0
		if self._startTime==0:
			self._startTime = time.time()
		else:
			timeDiff = time.time() - self._startTime
		return timeDiff
		
	##Reset effect
	##@param startValue : value when effect start
	##@param endValue : value when effect end
	def reset(self, startValue, endValue):
		self._startValue = startValue #b
		self._endValue = endValue #c
		self._startTime = 0
		if self._endValue>=self._startValue:
			self._valueToReach = self._endValue - self._startValue
		else:
			self._valueToReach = self._startValue - self._endValue
		print '---> RESET : startVal=%f, endVal=%f' % (self._startValue, self._endValue)

	##Return value to scroll (in pixel)
	def getScrollValue(self):
		raise VirtualWidgetException('getScrollValue() must be implemented!')

		
class NewKineticLinearEffect(NewKineticEffect):
	##Constructor
	##@param duration : duration of the effect
	def __init__(self, duration):
		#Super constructor
		NewKineticEffect.__init__(self, duration)
	
	##Return value to scroll (in pixel)
	def getScrollValue(self):
		#compute time diff
		timeDiff = self._getTimeDiff()
					
		#return scroll value
		if timeDiff<=1.0:
			if self._endValue>=self._startValue:
				return self._valueToReach * timeDiff / self._duration + self._startValue;
			else:
				return self._startValue - self._valueToReach * timeDiff / self._duration;
		else:
			#stop effect
			return 0
			

class NewKineticElasticEffect(NewKineticEffect):
	##Constructor
	##@param duration : duration of the effect
	def __init__(self, duration, period=0.25, amplitude=0):
		#set members
		self._period = period
		self._amplitude = amplitude
	
		#Super constructor
		NewKineticEffect.__init__(self, duration)
	
	##Return value to scroll (in pixel)
	def getScrollValue(self):
		#compute time diff
		timeDiff = self._getTimeDiff()
		
		#compute formula
		if timeDiff==0:
			return self._startValue
		timeDiff = timeDiff / self._duration
		if timeDiff==1:
			return self._startValue + self._valueToReach
		if not self._period:
			self._period = self._duration * 0.3
		if self._amplitude<abs(self._valueToReach):
			self._amplitude = self._valueToReach
			s = self._period/4
		else:
			s = self._period/(2*math.pi) * math.asin(self._valueToReach / self._amplitude)
		return self._amplitude*math.pow(2,-10*timeDiff) * math.sin( (timeDiff*self._duration-s)*(2*math.pi)/self._period ) + self._valueToReach + self._startValue;

			
class NewKineticEaseInQuadEffect(NewKineticEffect):
	##Constructor
	##@param duration : duration of the effect
	def __init__(self, duration):
		#Super constructor
		NewKineticEffect.__init__(self, duration)
	
	##Return value to scroll (in pixel)
	def getScrollValue(self):
		#compute time diff
		timeDiff = self._getTimeDiff()

		#return scroll value
		#return c*(t/=d)*t + b;
		if timeDiff<=1.0:
			if self._endValue>=self._startValue:
				timeDiff = timeDiff / self._duration
				return self._valueToReach * timeDiff * timeDiff + self._startValue;
			else:
				return self._startValue - self._valueToReach * timeDiff / self._duration;
		else:
			#stop effect
			return 0


##Base class for kinetic effect
class KineticEffect():
	##Constructor
	def __init__(self):
		#nothing to do
		pass
		
	##Reset effect
	def reset(self):
		raise VirtualWidgetException('reset() must be implemented!')
	
	##Apply kinetic effect and return value to scroll in kinetic widget
	##@param speed : speed applied by user (in pixel/s)
	##@return number of pixel to scroll
	def getScrollValue(self, speed):
		raise VirtualWidgetException('getScrollValue() must be implemented!')


		
##Class that describe kinetic linear effect
class KineticLinearEffect(KineticEffect):
	##Constructor
	def __init__(self):
		#super constructor
		KineticEffect.__init__(self)
		
		#set members
		self.__speed = 0.0
		self.__factor = 0.95
	
	##Reset effect
	def reset(self):
		self.__speed = 0.0
	
	##Apply kinetic effect
	def getScrollValue(self, speed):
		#start effect?
		if self.__speed==0.0:
			self.__speed = speed
	
		#compute effect
		self.__speed = self.__speed * self.__factor
		return self.__speed


		
##Class that describe exponential effect
class KineticExpEffect(KineticEffect):
	##Constructor
	##@param duration : effect duration
	def __init__(self):
		#super constructor
		KineticEffect.__init__(self)
		
		#set members
		self.__slope = 6.0
		self.__duration = 0.5
		self.__initTime = 0.0
		
	##Reset effect
	def reset(self):
		self.__initTime = 0.0
	
	##Apply kinetic effect
	def getScrollValue(self, speed):
		#get current time
		currentTime = time.time()
	
		#start effect?
		if self.__initTime==0.0:
			self.__initTime = currentTime
		
		#diff start time and current time
		diffTime = currentTime - self.__initTime
		
		#compute scroll value
		exp = 1.0 - math.exp( self.__slope * ( diffTime - self.__duration) )
		return speed * exp
		
		
		
##Class that describe arc tan effect
class KineticArcTanEffect(KineticEffect):
	##Constructor
	##@param duration : effect duration
	def __init__(self):
		#super constructor
		KineticEffect.__init__(self)
		
		#set members
		self.__initTime = 0.0
		
	##Reset effect
	def reset(self):
		self.__initTime = 0.0
	
	##Apply kinetic effect
	def getScrollValue(self, speed):
		#get current time
		currentTime = time.time()
	
		#start effect?
		if self.__initTime==0.0:
			self.__initTime = currentTime
		
		#diff start time and current time
		diffTime = currentTime - self.__initTime
		
		#compute scroll value
		factor = 2.607148
		exp = (-math.atan(7.0*diffTime-2.0)+1.5) / factor
		return float(speed) * exp
		
		
		
##Base class for kinetic widget
class KineticWidget(gtk.DrawingArea):
	##constructor
	def __init__(self, orientation, kineticEffect=None):
		#super constructor
		gtk.DrawingArea.__init__(self)
	
		#members
		#TODO sensitive : kinetic list can be controlled by software only (no user gesture permitted)
		self.__orientation = orientation
		self.__virtualList = KineticVirtualList(orientation)
		self.virtualRect = None
		self.__virtualPosition = geometry.Point()
		self.__virtualPositionAtClick = geometry.Point()
		self.__clickedPosition = None
		self.__speed = 0
		if self.__orientation==gtk.ORIENTATION_VERTICAL:
			self.__kineticDirection = gtk.DIR_DOWN
		else:
			self.__kineticDirection = gtk.DIR_RIGHT
		self.__stopScrolling = False
		self.__scrolling = False
		self.__redraw = False
		self.__adjustment = gtk.Adjustment()
		self.__pixmap = None
		self.__pixmap_selected = None
		if kineticEffect==None:
			#self.__kineticEffect = NewKineticEaseInQuadEffect(1)
			#self.__kineticEffect = NewKineticLinearEffect(1)
			self.__kineticEffect = NewKineticElasticEffect(1)
		
		#events
		self.add_events(gtk.gdk.BUTTON_PRESS_MASK | gtk.gdk.BUTTON_RELEASE_MASK | gtk.gdk.POINTER_MOTION_MASK)
		self.connect("expose_event", self.expose)
		self.connect('button-press-event', self.__clicked)
		self.connect('button-release-event', self.__released)
		self.connect('motion_notify_event', self.__moved)
		self.connect('realize', self.realize)
		self.connect('size_allocate', self.resize)
		
		#add specific timer to redraw widget
		gobject.timeout_add(10, self.redraw_canvas)
		
		#create custom signals
		gobject.signal_new('clicked', KineticWidget, gobject.SIGNAL_ACTION, gobject.TYPE_NONE, (KineticWidgetItem,))
		
	##Return adjustement (to use with scrollbar for example)
	def getAdjustment(self):
		return self.__adjustment
	
	##Action when cursor is moving above widget
	def __moved(self, widget, event):
		#get current position
		currentPosition = KineticClick(event)
		
		if self.__clickedPosition!=None:
			#init
			dist = 0
			
			if self.__orientation==gtk.ORIENTATION_VERTICAL:
				#compute distance between clicked point and current position
				dist = int(self.__clickedPosition.position.yDistance(currentPosition.position))
			
				#update virtual position
				self.__updateVirtualPosition(0, self.__virtualPositionAtClick.y + dist)
				
			else:
				#compute distance between clicked point and current position
				dist = int(self.__clickedPosition.position.xDistance(currentPosition.position))
			
				#update virtual position
				self.__updateVirtualPosition(self.__virtualPositionAtClick.x + dist, 0)
		
			#redraw
			self.__redraw = True
	
	##Action when mouse button is pressed
	def __clicked(self, widget, event):
		#stop scrolling if already running
		if self.__scrolling:
			self.__stopScrolling = True
			self.__scrolling = False
	
		#save clicked position
		self.__clickedPosition = KineticClick(event)
		
		#save virtual position at click to compute scrolling distance (see __moved)
		self.__virtualPositionAtClick = geometry.Point(self.virtualRect.x, self.virtualRect.y)
	
	##Action when mouse button is released 
	def __released(self, widget, event):
		#store current position
		releasedPosition = KineticClick(event)
		
		#detect if it's a click or a scroll
		if self.__orientation==gtk.ORIENTATION_VERTICAL:
			dist = self.__clickedPosition.position.yDistance(releasedPosition.position)
		else:
			dist = self.__clickedPosition.position.xDistance(releasedPosition.position)
			
		if abs(dist)<10:
			#it's an item selection
			
			#compute virtual position
			listPosition = geometry.Point(self.virtualRect.x + releasedPosition.position.x, self.virtualRect.y + releasedPosition.position.y)
			
			#get clicked item
			selectedItem = self.__virtualList.selectItem(listPosition)
			
			#redraw canvas if item has surface_selected
			if selectedItem!=None:
				#if selectedItem.pixmap_selected!=None:
				self.__redraw = True
			
				#launch event
				self.emit('clicked', selectedItem)
		else:
			#it's a kinetic scrolling
			'''self.__kineticEffect.reset(self.virtualRect.y, self.virtualRect.y+300)
			
			#reset flags
			self.__stopScrolling = False
		
			#compute speed in pixel/s
			time = releasedPosition.time - self.__clickedPosition.time
			speed = int((dist / time) / 10)
			if speed<0:
				if self.__orientation==gtk.ORIENTATION_VERTICAL:
					self.__kineticDirection = gtk.DIR_UP
				else:
					self.__kineticDirection = gtk.DIR_LEFT
			else:
				if self.__orientation==gtk.ORIENTATION_VERTICAL:
					self.__kineticDirection = gtk.DIR_DOWN
				else:
					self.__kineticDirection = gtk.DIR_RIGHT
			self.__speed = abs(speed)'''
			
			#reset flags
			self.__stopScrolling = False
			
			#compute speed in pixel/s
			time = releasedPosition.time - self.__clickedPosition.time
			speed = int((dist / time))
			print 'SPEED=%f' % speed
			
			#reset scrolling effect
			scrollBegin = 0
			scrollEnd = 0
			if speed<0:
				if self.__orientation==gtk.ORIENTATION_VERTICAL:
					self.__kineticDirection = gtk.DIR_UP
					scrollBegin = self.virtualRect.y
					scrollEnd = self.virtualRect.y + speed * self.__kineticEffect._duration
				else:
					self.__kineticDirection = gtk.DIR_LEFT
					self.__kineticEffect.reset(self.virtualRect.y, self.virtualRect.y+300)
			else:
				if self.__orientation==gtk.ORIENTATION_VERTICAL:
					self.__kineticDirection = gtk.DIR_DOWN
					scrollBegin = self.virtualRect.y
					scrollEnd = self.virtualRect.y + speed * self.__kineticEffect._duration
				else:
					self.__kineticDirection = gtk.DIR_RIGHT
					self.__kineticEffect.reset(self.virtualRect.y, self.virtualRect.y+300)
			if scrollEnd<0:
				scrollEnd = 0
			self.__kineticEffect.reset(scrollBegin, scrollEnd)
			
			#launch kinetic
			if time<=0.25:
				gobject.timeout_add(10, self.__kinetic)
			else:
				self.__speed = 0
		#reset
		self.__clickedPosition = None
	
	##Execute kinetic movement
	def __kinetic(self):
		if not self.__stopScrolling:
			#apply effect
			scroll = int(self.__kineticEffect.getScrollValue())
			print 'SCROLL=%d' % scroll
			
			if scroll==0.0:
				#stop scrolling
				self.__scrolling = False
				return False
			else:
				#update virtual position
				if self.__orientation==gtk.ORIENTATION_VERTICAL:
					if self.__kineticDirection==gtk.DIR_UP:
						#self.__updateVirtualPosition(0, self.virtualRect.y-scroll)
						self.__updateVirtualPosition(0, scroll)
					else:
						#self.__updateVirtualPosition(0, self.virtualRect.y+scroll)
						self.__updateVirtualPosition(0, scroll)
				else:
					if self.__kineticDirection==gtk.DIR_LEFT:
						self.__updateVirtualPosition(self.virtualRect.x-scroll, 0)
					else:
						self.__updateVirtualPosition(self.virtualRect.x+scroll, 0)
					
				#redraw canvas
				self.__redraw = True
				
				#continu scrolling
				self.__scrolling = True
				return True
				
		else:
			#set widget is not scrolling
			self.__scrolling = False
		
			#don't restart timer
			return False
	
	##Update internal adjustment
	def __updateAdjustment(self):
		if self.virtualRect!=None:
			if self.__orientation==gtk.ORIENTATION_VERTICAL:
				self.__adjustment.value = self.virtualRect.y
				self.__adjustment.upper = self.__virtualList.height
				self.__adjustment.page_size = self.virtualRect.height
			else:
				self.__adjustment.value = self.virtualRect.x
				self.__adjustment.upper = self.__virtualList.width
				self.__adjustment.page_size = self.virtualRect.width
	
	##Update virtual position
	def __updateVirtualPosition(self, positionX, positionY):
		#compute new virtual position
		if self.__orientation==gtk.ORIENTATION_VERTICAL:
			self.virtualRect.y = positionY
			
			#fix virtual position to stop scrolling at first and last item
			if self.virtualRect.y<0:
				self.virtualRect.y = 0
				self.__stopScrolling = True
			if (self.virtualRect.y+self.virtualRect.height)>self.__virtualList.height:
				self.virtualRect.y = self.__virtualList.height - self.virtualRect.height
				self.__stopScrolling = True
				
			#update adjustment
			self.__updateAdjustment()
			
		else:
			self.virtualRect.x = positionX
			
			#fix virtual position to stop scrolling at first and last item
			if self.virtualRect.x<0:
				self.virtualRect.x = 0
				self.__stopScrolling = True
			if (self.virtualRect.x+self.virtualRect.width)>self.__virtualList.width:
				self.virtualRect.x = self.__virtualList.width - self.virtualRect.width
				self.__stopScrolling = True
				
			#update adjustment
			self.__updateAdjustment()
			
	##Add item
	def addItem(self, item):
		#save item
		self.__virtualList.addItem(item)
		
		#update adjustment
		self.__updateAdjustment()
	
	##Remove item
	##TODO
	def removeItem(self, item):
		#remove item
		#self.__virtualList.removeItem(item)
		
		#update adjustment
		self.__updateAdjustment()
		
	##Go to previous item
	##TODO
	def goToPreviousItem(self):
		
		pass
		
	##Go to next item
	##TODO
	def goToNextItem(self):
		pass
		
	##Go to specified item
	##TODO
	def goToItem(self, itemToScroll):
		#get current item in the middle of the kinetic widget
		middlePosition = geometry.Point(0, 0)
		if self.__orientation==gtk.ORIENTATION_VERTICAL:
			middlePosition.x = self.__virtualRect.width / 2
			middlePosition.y = self.__virtualRect.y + self.__virtualRect.height / 2
		elif self.__orientation==gtk.ORIENTATION_HORIZONTAL:
			middlePosition.x = self.__virtualRect.x + self.__virtualRect.width / 2
			middlePosition.y = self.__virtualRect.height / 2
		middleItem = self.__virtualList.selectItem(middlePosition)
	
		if middleItem!=None:
			#item found
			#init
			direction = gtk.DIR_UP
			speed = 70 #70 pixel/s
		
			#detect direction and compute speed to scroll to
			if self.__orientation==gtk.ORIENTATION_VERTICAL:
				#direction
				if middleItem.rect.y<=itemToScroll.rect.y:
					direction = gtk.DIR_DOWN
				else:
					direction = gtk.DIR_UP
					
				#speed
				distance = abs(middleItem.rect.y - itemToScroll.rect.y)
			
			elif self.__orientation==gtk.ORIENTATION_HORIZONTAL:
				#direction
				if middleItem.rect.x<=itemToScroll.rect.x:
					direction = gtk.DIR_LEFT
				else:
					direction = gtk.DIR_RIGHT
			
				#speed
				distance = abs(middleItem.rect.x - itemToScroll.rect.x)
				self.__speed = 0
				
		direction = gtk.DIR_UP
		if self.__orientation==gtk.ORIENTATION_VERTICAL:
			if self.__virtualRect.y>=item.rect.y:
				direction = gtk.DIR_UP
			else:
				direction = gtk.DIR_DOWN

		elif self.__orientation==gtk.ORIENTATION_HORIZONTAL:
			if self.__virtualRect.x>=item.rect.x:	
				direction = gtk.DIR_LEFT
			else:
				direction = gtk.DIR_RIGHT
				
		#compute speed
		
		
		#launch kinetic
		
	
	##Create default item pixmaps
	##Can be overridden to change default kinetic pixmaps
	def _createItemPixmaps(self):
		if self.__orientation==gtk.ORIENTATION_VERTICAL:
			#create normal pixmap
			self.__pixmap = DefaultItemPixmap(self, self.virtualRect.width, self.__virtualList.itemSize[1], self.__orientation)
		
			#create selected pixmap
			self.__pixmap_selected = DefaultSelectedItemPixmap(self, self.virtualRect.width, self.__virtualList.itemSize[1], self.__orientation)
		elif self.__orientation==gtk.ORIENTATION_HORIZONTAL:
			#create normal pixmap
			self.__pixmap = DefaultItemPixmap(self, self.__virtualList.itemSize[0], self.virtualRect.height, self.__orientation)
		
			#create selected pixmap
			self.__pixmap_selected = DefaultSelectedItemPixmap(self, self.__virtualList.itemSize[0], self.virtualRect.height, self.__orientation)
			
	##Expose widget
	def expose(self, widget, event):
		#draw
		#try:
		self.window.begin_paint_rect( (event.area.x, event.area.y, event.area.width, event.area.height) )
		self.draw()
		self.window.end_paint()
		#except Exception, e:
		#	print 'EXCEPTION : %s' % e

		return False
		
	##Redraw canvas
	def redraw_canvas(self):
		#process all other events (extracted from kagu kinetic widget)
		while gtk.events_pending():
			gtk.main_iteration(block=True)   

		if self.__redraw:
			self.draw()
			self.__redraw = False
		
		return True
		
	##Widget is resized
	def resize(self, *args):
		if self.window!=None:
			#same behaviour than realize
			self.realize(args)
		
	##Widget is realized
	def realize(self, *args):
		#get widget rect
		rect = self.get_allocation()
			
		#set virtual rect
		self.virtualRect = geometry.Rectangle(rect.x, rect.y, rect.width, rect.height)
			
		#update adjustment
		self.__updateAdjustment()
			
		#create pixmaps
		self._createItemPixmaps()
			
	##Main drawing method
	def draw(self):
		if self.window:
			#draw items
			self.__drawItems()

	##Draw items that need to be drawn
	def __drawItems(self):
		#init
		items = self.__virtualList.getItems()
		itemFound = False
		selectedItem = self.__virtualList.getSelectedItem()
	
		if self.__pixmap==None:
			#pixmaps are not ready
			pass
		elif self.__orientation==gtk.ORIENTATION_VERTICAL:
			#compute index to start to search items (-1 to make sure to have all items)
			start = int(math.ceil( ( float(self.virtualRect.y)/float(self.__virtualList.height) ) * len(items) )) - 1
			if start<0:
				start = 0

			for index in range(start, len(items)):
				#get item
				item = items[index]
							
				#remove virtual vertical offset
				rect = geometry.Rectangle(item.rect.x, item.rect.y, item.rect.width, item.rect.height)
				rect.y -= self.virtualRect.y
		
				#check if item needs to be drawn
				if rect.y<=self.virtualRect.height:
					#draw item into pixmap
					kineticPixmap = None
					if selectedItem==None or selectedItem!=item:
						kineticPixmap = self.__pixmap
						item.draw(kineticPixmap)
					elif selectedItem==item:
						kineticPixmap = self.__pixmap_selected
						item.drawSelected(kineticPixmap)
						
					#draw item
					self.window.draw_drawable(self.get_style().fg_gc[gtk.STATE_NORMAL], kineticPixmap.pixmap, 0, 0, rect.x, rect.y, self.virtualRect.width, rect.height)
								
					#set flag
					itemFound = True
					
				elif itemFound:
					#item found and current item is outside specified rect, we can stop statement
					break

		elif self.__orientation==gtk.ORIENTATION_HORIZONTAL:
			#compute index to start to search items (-1 to make sure to have all items)
			start = int(math.ceil( ( float(self.virtualRect.x)/float(self.__virtualList.width) ) * len(items) )) - 1
			if start<0:
				start = 0
		
			for index in range(start, len(items)):
				#get item
				item = items[index]
							
				#remove virtual vertical offset
				rect = geometry.Rectangle(item.rect.x, item.rect.y, item.rect.width, item.rect.height)
				rect.x -= self.virtualRect.x
		
				#check if item needs to be drawn
				if rect.x<=self.virtualRect.width:
					#create cache if necessary
					#if item.pixmap==None:
					#	item.createCache(self.window)
			
					#draw item into pixmap
					kineticPixmap = None
					if selectedItem==None or selectedItem!=item:
						kineticPixmap = self.__pixmap
						item.draw(kineticPixmap)
					elif selectedItem==item:
						kineticPixmap = self.__pixmap_selected
						item.drawSelected(kineticPixmap)
						
					#draw item
					self.window.draw_drawable(self.get_style().fg_gc[gtk.STATE_NORMAL], kineticPixmap.pixmap, 0, 0, rect.x, rect.y, rect.width, self.virtualRect.height)
								
					#set flag
					itemFound = True
					
				elif itemFound:
					#item found and current item is outside specified rect, we can stop statement
					break

##Horizontal kinetic list with scrollbar
class KineticVList(gtk.HBox):
	def __init__(self, scrollbar=True):
		#super constructor
		gtk.HBox.__init__(self)
		
		#create kinetic widget
		self.kinetic = KineticWidget(gtk.ORIENTATION_VERTICAL)
		self.pack_start(self.kinetic, expand=True, fill=True, padding=0)
		
		#create hscrollbar
		if scrollbar:
			self.__scrollbar = gtk.VScrollbar(self.kinetic.getAdjustment())
			self.__eventBox = gtk.EventBox()
			self.__eventBox.set_above_child(True)
			self.__eventBox.add(self.__scrollbar)
			self.pack_start(self.__eventBox, expand=False, fill=True, padding=0)

##Vertical kinetic list with vertical scrollbar
class KineticHList(gtk.VBox):
	def __init__(self, scrollbar=True):
		#super constructor
		gtk.VBox.__init__(self)
		
		#create kinetic widget
		self.kinetic = KineticWidget(gtk.ORIENTATION_HORIZONTAL)
		self.pack_start(self.kinetic, expand=True, fill=True, padding=0)
		
		#create hscrollbar
		if scrollbar:
			self.__scrollbar = gtk.HScrollbar(self.kinetic.getAdjustment())
			self.__eventBox = gtk.EventBox()
			self.__eventBox.set_above_child(True)
			self.__eventBox.add(self.__scrollbar)
			self.pack_start(self.__eventBox, expand=False, fill=True, padding=0)
